"""
///////////////////////////////////////////////////////////////////////////////
//        Copyright (c) 2012-2020 Oscar Riveros. all rights reserved.        //
//                        oscar.riveros@peqnp.science                        //
//                                                                           //
//   without any restriction, Oscar Riveros reserved rights, patents and     //
//  commercialization of this knowledge or derived directly from this work.  //
///////////////////////////////////////////////////////////////////////////////
"""

__author__ = "Oscar Riveros"
__copyright__ = "Copyright (c) 2012-2019 Oscar Riveros. all rights reserved."
__credits__ = ["Oscar Riveros"]
__license__ = "PEQNP Licence"
__version__ = "0.0.1"
__maintainer__ = "Oscar Riveros"
__email__ = "oscar.riveros@peqnp-lib.science"
__status__ = "Production"

from .solver import hess_binary, hess_sequence
from .stdlib import *
